@@ -630,7 +630,11 @@ readline(FILE *fd, int want_command)
 			while (isblank(*p))
 				p++;
 
+#if ENABLE_FEATURE_MAKE_EXTENSIONS
+			if (*p != '\n' && (posix ? *str != '#' : *p != '#'))
+#else
 			if (*p != '\n' && *str != '#')
+#endif
 				return str;
 		}
 
@@ -639,9 +643,10 @@ readline(FILE *fd, int want_command)
 }
 
 /*
- * Return TRUE if the argument is a known suffix.
+ * Return a pointer to the suffix name if the argument is a known suffix
+ * or NULL if it isn't.
  */
-int
+const char *
 is_suffix(const char *s)
 {
 	struct name *np;
@@ -652,16 +657,54 @@ is_suffix(const char *s)
 	for (rp = np->n_rule; rp; rp = rp->r_next) {
 		for (dp = rp->r_dep; dp; dp = dp->d_next) {
 			if (strcmp(s, dp->d_name->n_name) == 0) {
-				return TRUE;
+				return dp->d_name->n_name;
+			}
+		}
+	}
+	return NULL;
+}
+
+#if ENABLE_FEATURE_MAKE_EXTENSIONS
+/*
+ * Return TRUE if the argument is formed by concatenating two
+ * known suffixes.
+ */
+int
+is_inference_target(const char *s)
+{
+	struct name *np;
+	struct rule *rp1, *rp2;
+	struct depend *dp1, *dp2;
+
+	np = newname(".SUFFIXES");
+	for (rp1 = np->n_rule; rp1; rp1 = rp1->r_next) {
+		for (dp1 = rp1->r_dep; dp1; dp1 = dp1->d_next) {
+			const char *suff1 = dp1->d_name->n_name;
+			size_t len = strlen(suff1);
+
+			if (strncmp(s, suff1, len) == 0) {
+				for (rp2 = np->n_rule; rp2; rp2 = rp2->r_next) {
+					for (dp2 = rp2->r_dep; dp2; dp2 = dp2->d_next) {
+						const char *suff2 = dp2->d_name->n_name;
+						if (strcmp(s + len, suff2) == 0) {
+							return TRUE;
+						}
+					}
+				}
 			}
 		}
 	}
 	return FALSE;
 }
+#endif
 
-#define T_NORMAL 0
-#define T_SPECIAL 1
-#define T_INFERENCE 2
+enum {
+	T_NORMAL    =  0,
+	T_SPECIAL   = (1 << 0),
+	T_INFERENCE = (1 << 1), // Inference rule
+	T_NOPREREQ  = (1 << 2), // If set must not have prerequisites
+	T_COMMAND   = (1 << 3), // If set must have commands, if unset must not
+};
 
 /*
  * Determine if the argument is a special target and return a set
@@ -670,7 +713,6 @@ is_suffix(const char *s)
 static int
 target_type(char *s)
 {
-	char *sfx;
 	int ret;
 	static const char *s_name[] = {
 		".DEFAULT",
@@ -689,26 +731,50 @@ target_type(char *s)
 #endif
 	};
 
-	if (*s != '.')
-		return T_NORMAL;
+	static const uint8_t s_type[] = {
+		T_SPECIAL | T_NOPREREQ | T_COMMAND,
+		T_SPECIAL | T_NOPREREQ,
+		T_SPECIAL,
+		T_SPECIAL,
+		T_SPECIAL,
+		T_SPECIAL,
+#if ENABLE_FEATURE_MAKE_POSIX_2024
+		T_SPECIAL,
+		T_SPECIAL | T_NOPREREQ,
+		T_SPECIAL | T_NOPREREQ,
+#endif
+#if ENABLE_FEATURE_MAKE_EXTENSIONS
+		T_SPECIAL,
+#endif
+	};
 
 	// Check for one of the known special targets
 	for (ret = 0; ret < sizeof(s_name)/sizeof(s_name[0]); ret++)
 		if (strcmp(s_name[ret], s) == 0)
-			return T_SPECIAL;
+			return s_type[ret];
 
 	// Check for an inference rule
 	ret = T_NORMAL;
-	sfx = suffix(s);
-	if (is_suffix(sfx)) {
-		if (s == sfx) {	// Single suffix rule
-			ret = T_INFERENCE;
-		} else {
-			// Suffix is valid, check that prefix is too
-			*sfx = '\0';
-			if (is_suffix(s))
-				ret = T_INFERENCE;
-			*sfx = '.';
+#if ENABLE_FEATURE_MAKE_EXTENSIONS
+	if (!posix) {
+		if (is_suffix(s) || is_inference_target(s)) {
+			ret = T_INFERENCE | T_NOPREREQ | T_COMMAND;
+		}
+	} else
+#endif
+	{
+		// In POSIX inference rule targets must contain one or two dots
+		char *sfx = suffix(s);
+		if (*s == '.' && is_suffix(sfx)) {
+			if (s == sfx) {	// Single suffix rule
+				ret = T_INFERENCE | T_NOPREREQ | T_COMMAND;
+			} else {
+				// Suffix is valid, check that prefix is too
+				*sfx = '\0';
+				if (is_suffix(s))
+					ret = T_INFERENCE | T_NOPREREQ | T_COMMAND;
+				*sfx = '.';
+			}
 		}
 	}
 	return ret;
@@ -1186,7 +1252,7 @@ input(FILE *fd, int ilevel)
 			free(copy3);
 			*s = '\0';
 		}
-		semicolon_cmd = cp != NULL;
+		semicolon_cmd = cp != NULL && cp->c_cmd[0] != '\0';
 
 		// Create list of prerequisites
 		dp = NULL;
@@ -1281,13 +1347,32 @@ input(FILE *fd, int ilevel)
 
 				np = newname(p);
 				if (ttype != T_NORMAL) {
-					if (ttype == T_INFERENCE
-							IF_FEATURE_MAKE_EXTENSIONS(&& posix)) {
-						if (semicolon_cmd)
-							error_in_inference_rule("'; command'");
-						seen_inference = TRUE;
+					// Enforce prerequisites/commands in POSIX mode
+					if (IF_FEATURE_MAKE_EXTENSIONS(posix &&) 1) {
+						if ((ttype & T_NOPREREQ) && dp)
+							error_not_allowed("prerequisites", p);
+						if ((ttype & T_INFERENCE)) {
+							if (semicolon_cmd)
+								error_in_inference_rule("'; command'");
+							seen_inference = TRUE;
+						}
+						if ((ttype & T_COMMAND) && !cp &&
+								!((ttype & T_INFERENCE) && !semicolon_cmd))
+							error("commands required for %s", p);
+						if (!(ttype & T_COMMAND) && cp)
+							error_not_allowed("commands", p);
+					}
+
+					if ((ttype & T_INFERENCE)) {
+						np->n_flag |= N_INFERENCE;
+#if ENABLE_FEATURE_MAKE_EXTENSIONS
+					} else if (strcmp(p, ".DEFAULT") == 0) {
+						// .DEFAULT rule is a special case
+						np->n_flag |= N_SPECIAL | N_INFERENCE;
+#endif
+					} else {
+						np->n_flag |= N_SPECIAL;
 					}
-					np->n_flag |= N_SPECIAL;
 				} else if (!firstname) {
 					firstname = np;
 				}
@@ -1300,7 +1385,7 @@ input(FILE *fd, int ilevel)
 				globfree(&gd);
 #endif
 		}
-		if (seen_inference && count != 1)
+		if (IF_FEATURE_MAKE_EXTENSIONS(posix &&) seen_inference && count != 1)
 			error_in_inference_rule("multiple targets");
 
 		// Prerequisites and commands will be unused if there were
